"""
Routines for solving the nonlinear part of the Grad-Shafranov equation

Copyright 2016-2019 Ben Dudson, University of York. Email: benjamin.dudson@york.ac.uk

This file is part of FreeGS.

FreeGS is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FreeGS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with FreeGS.  If not, see <http://www.gnu.org/licenses/>.
"""

from numpy import amin, amax, array
from . import critical
import numpy as np


def solve(
    eq,
    profiles,
    constrain=None,
    rtol=1e-3,
    atol=1e-10,
    blend=0.0,
    show=False,
    axis=None,
    pause=0.0001,
    psi_bndry=None,
    maxits=50,
    convergenceInfo=False,
    check_limited=False,
    wait_for_limited=False,
    limit_it=0,
):
    """
    Perform Picard iteration to find solution to the Grad-Shafranov equation

    Parameters
    ----------
    eq :
        an Equilibrium object (equilibrium.py)
    profiles :
        A Profile object for toroidal current (jtor.py)

    rtol:
        Relative tolerance (change in psi)/( max(psi) - min(psi) )
    atol:
        Absolute tolerance, change in psi
    blend:
        Blending of previous and next psi solution

            psi{n+1} <- psi{n+1} * (1-blend) + blend * psi{n}

    show:
        If true, plot the plasma equilibrium at each nonlinear step
    axis:
        Specify a figure to plot onto. Default (None) creates a new figure
    pause:
        Delay between output plots. If negative, waits for window to be closed
    maxits:
        Maximum number of iterations. Set to None for no limit.
        If this limit is exceeded then a RuntimeError is raised.
    convergenceInfo:
        True/False toggle for outputting convergence data.
    check_limited:
        True/False toggle to control checking for limited plasmas.
    wait_for_limited:
        True/False toggle to keep iterating until the plasma is limited.
    limit_it:
        Sometimes waiting some number of interations before checking
        if a plasma is limited can improve convergence. This sets the
        number of iterations to wait.
    """

    if constrain is not None:
        # Set the coil currents to get X-points in desired locations
        constrain(eq)

    # Get the total psi = plasma + coils
    psi = eq.psi()

    if show:
        import matplotlib.pyplot as plt
        from .plotting import plotEquilibrium

        if pause > 0.0 and axis is None:
            # No axis specified, so create a new figure
            fig = plt.figure()
            axis = fig.add_subplot(111)

    # Count number of iterations
    iteration = 0

    # Initial relative change in psi (set high to prevent immediate convergence)
    psi_relchange = 10.0

    # Initial psi_bndry (set low to prevent immediate convergence)
    bndry = 0.0

    # Set an initial value for bndry_change (set high to prevent immediate convergence)
    bndry_change = np.inf

    # Plasma assumed to not be limited at first
    has_been_limited = False

    # It is not yet ok to stop itterating
    ok_to_break = False

    psi_maxchange_iterations, psi_relchange_iterations = [], []

    # Start main loop
    while True:

        if show:
            # Plot state of plasma equilibrium
            if pause < 0:
                fig = plt.figure()
                axis = fig.add_subplot(111)
            else:
                axis.clear()

            plotEquilibrium(eq, axis=axis, show=False)

            if pause < 0:
                # Wait for user to close the window
                plt.show()
            else:
                # Update the canvas and pause
                # Note, a short pause is needed to force drawing update
                axis.figure.canvas.draw()
                plt.pause(pause)

        # Copy psi to compare at the end
        psi_last = psi.copy()

        # Boundary flux can also be used as a convergence criterion, so note it
        bndry_last = bndry

        if (iteration >= limit_it or has_been_limited) and check_limited:
            # The user wishes to check for a limited plasma.
            # The minimum number or iterations has passed.
            # If it is ever found to be limited, keep checking for
            # further limited plasmas.

            eq.check_limited = True
            eq.solve(profiles, psi=psi, psi_bndry=eq.psi_bndry)

        else:

            # Either the user does not wish to check for a limited plasma,
            # or not enough iterations have passed yet.
            eq.check_limited = False
            eq.solve(profiles, psi=psi, psi_bndry=psi_bndry)

        # Keep track of whether or not the plasma has at all been limited.
        if eq.is_limited:
            has_been_limited = True

        # If the equilibrium is limited, is must remain so for atleast
        # 1 iteration to allow sudden diverted->limited changes
        # to propagate to the plasma internal profiles. This is captured
        # by also checking if psi_bndry converges.

        if eq.psi_bndry is not None:
            # Check is psi_bndry converges
            bndry = eq.psi_bndry
            bndry_change = bndry_last - bndry
            bndry_relchange = abs(bndry_change / bndry)

        else:
            # Dummy condition to prevent boundary
            # convergence when there is no boundary
            # ie set the change to > rtol
            bndry_relchange = 2.0 * rtol

        # Get the new psi, including coils
        psi = eq.psi()

        # Compare against last solution
        psi_change = psi_last - psi
        psi_maxchange = amax(abs(psi_change))
        psi_relchange = psi_maxchange / (amax(psi) - amin(psi))

        psi_maxchange_iterations.append(psi_maxchange)
        psi_relchange_iterations.append(psi_relchange)

        # User has the option to keep converging until limited
        if not wait_for_limited:
            # User does not wish to wait for the plasma to become limited
            ok_to_break = True

        elif wait_for_limited and eq.is_limited:
            # User wants to check if plasma limited and it is actually limited
            ok_to_break = True

        else:
            # The user wants to wait for a limited plasma. The plasma is not limited.
            ok_to_break = False

        if show:
            print("psi_relchange: " + str(psi_relchange))
            print("bndry_relchange: " + str(bndry_relchange))
            print("bndry_change: " + str(bndry_change))
            print("\n")

        # Check if the changes in psi are small enough and that it is ok to start checking for convergence
        if (
            ((psi_maxchange < atol) or (psi_relchange < rtol))
            and ((bndry_relchange < rtol) or (abs(bndry_change) < atol))
            and ok_to_break
        ):
            break

        # Adjust the coil currents
        if constrain is not None:
            constrain(eq)

        psi = (1.0 - blend) * eq.psi() + blend * psi_last

        # Check if the maximum iterations has been exceeded
        iteration += 1
        if maxits and iteration > maxits:
            raise RuntimeError(
                "Picard iteration failed to converge (too many iterations)"
            )
    if convergenceInfo:
        return array(psi_maxchange_iterations), array(psi_relchange_iterations)



"""
Plotting using matplotlib

Copyright 2016 Ben Dudson, University of York. Email: benjamin.dudson@york.ac.uk

This file is part of FreeGS.

FreeGS is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FreeGS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with FreeGS.  If not, see <http://www.gnu.org/licenses/>.

"""

from numpy import linspace, amin, amax
from . import critical


def plotCoils(coils, axis=None):
    import matplotlib.pyplot as plt

    if axis is None:
        fig = plt.figure()
        axis = fig.add_subplot(111)

    return axis


def plotConstraints(control, axis=None, show=True):
    """
    Plots constraints used for coil current control

    axis     - Specify the axis on which to plot
    show     - Call matplotlib.pyplot.show() before returning

    """

    import matplotlib.pyplot as plt

    if axis is None:
        fig = plt.figure()
        axis = fig.add_subplot(111)

    # Locations of the X-points
    for r, z in control.xpoints:
        axis.plot(r, z, "bx")

    if control.xpoints:
        axis.plot([], [], "bx", label="X-point constraints")

    # Isoflux surfaces
    for r1, z1, r2, z2 in control.isoflux:
        axis.plot([r1, r2], [z1, z2], ":b^")

    if control.isoflux:
        axis.plot([], [], ":b^", label="Isoflux constraints")

    if show:
        plt.legend()
        plt.show()

    return axis

def plotEquilibrium(eq, axis=None, show=True, oxpoints=True, wall=True, plot_sensors=True):
    """
    Plot the equilibrium flux surfaces

    axis     - Specify the axis on which to plot
    show     - Call matplotlib.pyplot.show() before returning
    oxpoints - Plot X points as red circles, O points as green circles
    wall     - Plot the wall (limiter)
    sensors  - Plot the sensors
    """

    import matplotlib.pyplot as plt

    R = eq.R
    Z = eq.Z
    psi = eq.psi()

    if axis is None:
        fig = plt.figure()
        axis = fig.add_subplot(111)

    levels = linspace(amin(psi), amax(psi), 100)

    axis.contour(R, Z, psi, levels=levels)
    axis.set_aspect("equal")
    axis.set_xlabel("Major radius [m]")
    axis.set_ylabel("Height [m]")

    if oxpoints:
        # Add O- and X-points
        opt, xpt = critical.find_critical(eq.R, eq.Z, psi)

        for r, z, _ in xpt:
            axis.plot(r, z, "rx")
        for r, z, _ in opt:
            axis.plot(r, z, "go")

        if xpt:
            psi_bndry = eq.psi_bndry  # xpt[0][2]
            axis.contour(eq.R, eq.Z, psi, levels=[psi_bndry], colors="r")

            # Add legend
            axis.plot([], [], "rx", label="X-points")
            axis.plot([], [], "r", label="Separatrix")
        if opt:
            axis.plot([], [], "go", label="O-points")

    if wall and eq.tokamak.wall and len(eq.tokamak.wall.R):
        axis.plot(
            list(eq.tokamak.wall.R) + [eq.tokamak.wall.R[0]],
            list(eq.tokamak.wall.Z) + [eq.tokamak.wall.Z[0]],
            "k",
        )

    if plot_sensors:
        if eq.tokamak.sensors is not None :
            for sensor in eq.tokamak.sensors:
                sensor.plot(axis)
            axis.plot([], [], 'b', label='Rogowski Coil')
            axis.plot([], [], 'mo', label='Poloidal Field Sensor')
            axis.plot([], [], 'ro', label='Flux Loop Sensor')

    if show:
        plt.legend()
        plt.show()

    return axis



from freegs import geqdsk
from freegs import machine
from freegs.plotting import plotEquilibrium

tokamak = machine.TestTokamak()

with open("lsn.geqdsk") as f:
    eq = geqdsk.read(f, tokamak, show=True)

plotEquilibrium(eq)



#!/usr/bin/env python

import freegs
import matplotlib.pyplot as plt

#########################################
# Create the machine, which specifies coil locations
# and equilibrium, specifying the domain to solve over

tokamak = freegs.machine.DIIID()

eq = freegs.Equilibrium(tokamak=tokamak,
                        Rmin=0.1, Rmax=2.8,    # Radial domain
                        Zmin=-1.8, Zmax=1.8,   # Height range
                        nx=129, ny=129)          # Number of grid points

#########################################
# Plasma profiles

profiles = freegs.jtor.ConstrainPaxisIp(eq,
                                        159811, # Plasma pressure on axis [Pascals]
                                        -1533632, # Plasma current [Amps]
                                        -3.231962138124) # vacuum f = R*Bt

#########################################
# Coil current constraints
#
# Specify locations of the X-points
# to use to constrain coil currents

xpoints = [(1.285, -1.176),   # (R,Z) locations of X-points
           (1.2, 1.0)]

isoflux = [(1.285, -1.176, 1.2 ,1.2)] # (R1,Z1, R2,Z2) pair of locations

constrain = freegs.control.constrain(xpoints=xpoints, gamma=1e-12, isoflux=isoflux)

constrain(eq)

#########################################
# Nonlinear solve

freegs.solve(eq,          # The equilibrium to adjust
             profiles,    # The plasma profiles
             constrain,   # Plasma control constraints
             show=True)   # Shows results at each nonlinear iteration

# eq now contains the solution

print("Done!")

print("Plasma current: %e Amps" % (eq.plasmaCurrent()))
print("Pressure on axis: %e Pascals" % (eq.pressure(0.0)))
print("Plasma poloidal beta: %e" % (eq.poloidalBeta()))
print("Plasma volume: %e m^3" % (eq.plasmaVolume()))

eq.tokamak.printCurrents()

# plot equilibrium
axis = eq.plot(show=False)
tokamak.plot(axis=axis, show=False)
constrain.plot(axis=axis, show=True)

# Safety factor
plt.plot(*eq.q())
plt.xlabel(r"Normalised $\psi$")
plt.ylabel("Safety factor")
plt.grid()
plt.show()

##############################################
# Save to geqdsk file

from freegs import geqdsk

with open("diiid.geqdsk", "w") as f:
    geqdsk.write(eq, f)




import freegs
from freegs import geqdsk
from freegs.equilibrium import refine
from freegs.plotting import plotEquilibrium

# Reading MAST equilibrium, up-down symmetric coils
tokamak = freegs.machine.MAST()

#with open("g014220.00200") as f:
with open("mast.geqdsk") as f:
  eq = geqdsk.read(f, tokamak, show=True)

# Increase resolution by a factor of 2
eq2 = refine(eq)

# Re-solve, keeping the same control points and profiles
freegs.solve(eq2, eq._profiles, eq.control, rtol=1e-6)

# Save to G-EQDSK
with open("mast-highres.geqdsk", "w") as f:
  geqdsk.write(eq2, f)

plotEquilibrium(eq2)




"""
Classes and routines to represent coils and circuits

License
-------

Copyright 2022 Chris Marsden, Tokamak Energy. Email: chris.marsden@tokamakenergy.co.uk

This file is part of FreeGS.

FreeGS is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FreeGS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with FreeGS.  If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np
from scipy.interpolate import RectBivariateSpline
from .coil import Coil, AreaCurrentLimit
from .gradshafranov import Greens, GreensBr, GreensBz
from . import polygons


class PreCalcCoil(Coil):
    """
    This class represents a coil whose Green's functions have
    already been calculated by some external code. This is useful
    in modelling coils whose internal structure may be complex and
    whose current distribution may be highly non-uniform.

    The user needs to supply information on the R,Z grids that the
    Green's functions have been pre-calculated on, as well as the
    Br, Bz and psi data on said R,Z grid.

    public members
    --------------

    R, Z    - Location of the coil
    current - current in the coil in Amps
    turns   - Number of turns
    control - enable or disable control system
    area    - Cross-section area in m^2

    The total toroidal current carried by the coil is current * turns
    """

    # A dtype for converting to Numpy array and storing in HDF5 files
    dtype = np.dtype(
        [
            (str("RZlen"), int),  # Length of the R and Z arrays
            (str("R"), "10f8"),  # Note: Up to 10 points
            (str("Z"), "10f8"),  # Note: Up to 10 points
            (str("current"), np.float64),
            (str("turns"), int),
            (str("control"), bool),
        ]
    )

    def __init__(
        self,
        shape,
        Rgrid,
        Zgrid,
        mapBr,
        mapBz,
        mapPsi,
        current=0.0,
        turns=1,
        control=True,
    ):
        """
        Inputs
        ------
        shape  - outline of the coil shape as a list of points [(r1,z1), (r2,z2), ...]
                 Must have more than two points

        Rgrid   - 1D array of R coords that maps are calculated on.
        Zgrid   - 1D array of Z coords that maps are calculated on.
        mapBr   - 1D array of Br calculated on Rgrid,Zgrid for the coil @ 1A-turn.
        mapBz   - 1D array of Bz calculated on Rgrid,Zgrid for the coil @ 1A-turn.
        mapPsi  - 1D array of Psi calculated on Rgrid,Zgrid for the coil @ 1A-turn.
        current - The current in the circuit. The total current is current * turns.
        turns   - Number of turns in point coil(s) block. Total block current is current * turns.
        control - Enable or disable control system.

        """
        assert len(shape) > 2

        # Find the geometric middle of the coil
        # The R,Z properties have accessor functions to handle modifications
        self._R_centre = sum(r for r, z in shape) / len(shape)
        self._Z_centre = sum(z for r, z in shape) / len(shape)

        self.current = current
        self.turns = turns
        self.control = control
        self._area = abs(polygons.area(shape))
        self.shape = shape
        self._points = np.array([(r, z) for r, z in self.shape])

        # Data for the pre-calculated Green's functions
        self.Rgrid = np.transpose(Rgrid)[:, 0]
        self.Zgrid = np.transpose(Zgrid)[0, :]
        self.mapPsi = np.transpose(np.asarray(mapPsi))
        self.mapBr = np.transpose(np.asarray(mapBr))
        self.mapBz = np.transpose(np.asarray(mapBz))

        # Interpolators for the pre-calculated Green's functions
        self.cPsi = RectBivariateSpline(self.Rgrid, self.Zgrid, self.mapPsi)
        self.cBr = RectBivariateSpline(self.Rgrid, self.Zgrid, self.mapBr)
        self.cBz = RectBivariateSpline(self.Rgrid, self.Zgrid, self.mapBz)

    def controlPsi(self, R, Z):
        """
        Calculate poloidal flux at (R,Z) due to a unit current.
        """

        if isinstance(R, float) or isinstance(R, int):
            result = self.cPsi(R, Z)[0][0]
        else:
            result = self.cPsi(R, Z, grid=False)

        return result

    def controlBr(self, R, Z):
        """
        Calculate radial magnetic field Br at (R,Z) due to a unit current.
        """

        if isinstance(R, float) or isinstance(R, int):
            result = self.cBr(R, Z)[0][0]
        else:
            result = self.cBr(R, Z, grid=False)

        return result

    def controlBz(self, R, Z):
        """
        Calculate vertical magnetic field Br at (R,Z) due to a unit current.
        """

        if isinstance(R, float) or isinstance(R, int):
            result = self.cBz(R, Z)[0][0]
        else:
            result = self.cBz(R, Z, grid=False)

        return result

    def __repr__(self):
        return "PreCalcCoil({0}, current={1:.1f}, turns={2}, control={3})".format(
            self.shape, self.current, self.turns, self.control
        )

    @property
    def R(self):
        """
        Major radius of the coil in m
        """
        return self._R_centre

    @R.setter
    def R(self, Rnew):
        # Need to shift all points
        Rshift = Rnew - self._R_centre
        self._points = [(r + Rshift, z) for r, z in self._points]
        self._R_centre = Rnew

    @property
    def Z(self):
        """
        Height of the coil in m
        """
        return self._Z_centre

    @Z.setter
    def Z(self, Znew):
        # Need to shift all points
        Zshift = Znew - self._Z_centre
        self._points = [(r, z + Zshift) for r, z in self._points]
        self._Z_centre = Znew

    @property
    def area(self):
        return self._area

    @area.setter
    def area(self, area):
        raise ValueError("Area of a PreCalcCoil is fixed")

    def plot(self, axis=None, show=False):
        """
        Plot the coil shape, using axis if given
        """
        import matplotlib.pyplot as plt

        if axis is None:
            fig = plt.figure()
            axis = fig.add_subplot(111)

        r = [r for r, z in self.shape]
        z = [z for r, z in self.shape]
        axis.fill(r, z, color="gray")
        axis.plot(r, z, color="black")

        # Quadrature points
        # rquad = [r for r,z,w in self._points]
        # zquad = [z for r,z,w in self._points]
        # axis.plot(rquad, zquad, 'ro')

        return axis

    def to_numpy_array(self):
        """
        Helper method for writing output
        """
        RZlen = len(self.shape)
        R = np.zeros(10)
        Z = np.zeros(10)
        R[:RZlen] = [R for R, Z in self.shape]
        Z[:RZlen] = [Z for R, Z in self.shape]

        return np.array(
            (RZlen, R, Z, self.current, self.turns, self.control),
            dtype=self.dtype,
        )

    def __ne__(self, other):
        return not self == other



from .filament_coil import FilamentCoil
from .coil import Coil
from .machine import Circuit

import numpy as np


def test_single():
    """
    Test that a single turn FilamentCoil is the same as a Coil
    """

    fcoil = FilamentCoil(1.1, 0.2, current=100.0)
    coil = Coil(1.1, 0.2, current=100.0)

    assert np.isclose(coil.controlPsi(0.3, 0.1), fcoil.controlPsi(0.3, 0.1))

    assert np.isclose(coil.controlBr(0.3, 0.1), fcoil.controlBr(0.3, 0.1))


def test_two_filaments():
    """
    FilamentCoil with two filaments same as circuit with two coils at half current
    """

    fcoil = FilamentCoil([1.1, 1.2], [0.2, -0.3])

    circuit = Circuit(
        [
            ("A", FilamentCoil(1.1, 0.2), 0.5),
            ("B", FilamentCoil(1.2, -0.3), 0.5),
        ]
    )

    # Check the Greens functions (field produced by unit circuit current)
    assert np.isclose(circuit.controlPsi(0.3, 0.1), fcoil.controlPsi(0.3, 0.1))
    assert np.isclose(circuit.controlBr(0.3, 0.1), fcoil.controlBr(0.3, 0.1))
    assert np.isclose(circuit.controlBz(0.3, 0.1), fcoil.controlBz(0.3, 0.1))

    # Check the field, putting the same current
    fcoil.current = 100
    circuit.current = 100
    assert np.isclose(circuit.psi(0.3, 0.1), fcoil.psi(0.3, 0.1))
    assert np.isclose(circuit.Br(0.3, 0.1), fcoil.Br(0.3, 0.1))
    assert np.isclose(circuit.Bz(0.3, 0.1), fcoil.Bz(0.3, 0.1))


def test_two_turns():
    """
    FilamentCoil with two turns same as circuit with two coils
    """

    fcoil = FilamentCoil(1.1, 1.2, turns=2)

    circuit = Circuit(
        [
            ("A", FilamentCoil(1.1, 1.2), 1.0),
            ("B", FilamentCoil(1.1, 1.2), 1.0),
        ]
    )

    # Check the Greens functions (field produced by unit circuit current)
    assert np.isclose(circuit.controlPsi(0.3, 0.1), fcoil.controlPsi(0.3, 0.1))
    assert np.isclose(circuit.controlBr(0.3, 0.1), fcoil.controlBr(0.3, 0.1))
    assert np.isclose(circuit.controlBz(0.3, 0.1), fcoil.controlBz(0.3, 0.1))


def test_mirrored():
    """
    Mirrored FilamentCoil the same as two coils in a circuit at half current

    Filaments in a FilamentCoil are wired in parallel, not in serial.
    """
    fcoil = FilamentCoil([1.1, 1.1], [0.2, -0.2])

    circuit = Circuit(
        [
            ("A", Coil(1.1, 0.2), 0.5),
            ("B", Coil(1.1, -0.2), 0.5),
        ]
    )

    # Check
    assert np.isclose(circuit.controlPsi(0.3, 0.1), fcoil.controlPsi(0.3, 0.1))
    assert np.isclose(circuit.controlBr(0.3, 0.1), fcoil.controlBr(0.3, 0.1))
    assert np.isclose(circuit.controlBz(0.3, 0.1), fcoil.controlBz(0.3, 0.1))


def test_move_R():
    """
    Changing major radius property R changes all filament locations
    """

    dR = 0.6
    coil1 = FilamentCoil([1.1, 0.2], [1.2, -0.3], current=100.0)
    coil2 = FilamentCoil([1.1 + dR, 0.2 + dR], [1.2, -0.3], current=100.0)

    # Shift coil1 to same location as coil2
    coil1.R += dR

    assert np.isclose(coil1.controlPsi(0.4, 0.5), coil2.controlPsi(0.4, 0.5))

    assert np.isclose(coil1.controlBr(0.3, -0.2), coil2.controlBr(0.3, -0.2))

    assert np.isclose(coil1.controlBz(1.75, 1.2), coil2.controlBz(1.75, 1.2))


def test_move_Z():
    """
    Changing height property Z changes all filament locations
    """

    dZ = 0.4
    coil1 = FilamentCoil([1.1, 0.2], [1.2, -0.3], current=100.0)
    coil2 = FilamentCoil([1.1, 0.2], [1.2 + dZ, -0.3 + dZ], current=100.0)

    # Shift coil1 to same location as coil2
    coil1.Z += dZ

    assert np.isclose(coil1.controlPsi(0.4, 0.5), coil2.controlPsi(0.4, 0.5))

    assert np.isclose(coil1.controlBr(0.3, -0.2), coil2.controlBr(0.3, -0.2))

    assert np.isclose(coil1.controlBz(1.75, 1.2), coil2.controlBz(1.75, 1.2))
